/*
 * 1-dim non-linear Wave equation
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>
#include "bmp.c"

#define PI 3.14159265358979323846

/* global staff */
#define IMAX 500
#define NBUF 3
double v[NBUF][IMAX+1];
double u[NBUF][IMAX+1];
double du[IMAX], dv[IMAX], ds[IMAX];

/* window size */
int width = IMAX, height = width/2;

/* solving equation */
void Wave(void);
void (*pde)(void) = Wave;

/* global parameters */
#define L   (1.0)   /* length */
double k = 30;      /* elasticity parameter */

/* drawing parameters */
#define draw_step  ( (k > 3) ? int(k) : 3 )
double ymag  = 100.0;

/* flags */
int initialize = 1;
int ivalue= 0;
int pause  = 0;
int dump_images = 0;

/* for violin */
int violin = 0;
#define v0  (.003) // velocity of "yumi"

/*****************************************
 * solve Wave equation
 ----------------------------------------*/
void Wave(void)
//
#define dx   ( L/IMAX )
#define dt   ( dx/k )
#define x(i) ( dx*(i) )
#define n(D) ( (n+(D)+NBUF) % NBUF )
#define sch1 ( 1./cosh(30*(x(i) - .5)) )
#define sch2 ( 1./cosh(30*(x(i) - .6)) )
{
    static unsigned int n = 0;

    /* initial value */
    if (initialize == 1) {
	n = 0; initialize = 0; violin = 0;
	switch (ivalue) {
	case 'g':
	    for (int i = 1; i < IMAX; i++) {
		static double x0 = 1./6, vmax = .01;
		u[n(0)][i] = u[n(-1)][i] = 
		    - (k-1)/k/2*(1/x0-1/(1-x0))*vmax*vmax
			*((x(i)<x0) ? x(i)/x0 : (1-x(i))/(1-x0));
		v[n(0)][i] = v[n(-1)][i] = 
		    -vmax*((x(i)<x0) ? x(i)/x0 : (1-x(i))/(1-x0));
	    }
	    break;
	case 'p':
	    for (int i = 1; i < IMAX; i++) {
		static double x0 = 1./8;
		u[n(0)][i] = u[n(-1)][i] = 0;
		v[n(0)][i] = 0;
		v[n(-1)][i] = -0.3*dt*(1/cosh(60*(x(i) - x0)));
	    }
	    break;
	case 'v':
	    for (int i = 1; i < IMAX; i++) {
		u[n(0)][i] = u[n(-1)][i] = 0;
		v[n(0)][i] = v[n(-1)][i] = 0;
	    }
	    violin = 1;
	    break;
	default:
	    break;
	}
    }

    /* evolve */
    du[0] = (u[n(0)][1] - u[n(0)][0])/dx;
    dv[0] = (v[n(0)][1] - v[n(0)][0])/dx;
    ds[0] = sqrt((1+du[0])*(1+du[0]) + dv[0]*dv[0]);
    for (int i = 1; i < IMAX; i++) {
	du[i] = (u[n(0)][i+1] - u[n(0)][i])/dx;
	dv[i] = (v[n(0)][i+1] - v[n(0)][i])/dx;
	ds[i] = sqrt((1+du[i])*(1+du[i]) + dv[i]*dv[i]);
	u[n(+1)][i] = 2*u[n(0)][i] - u[n(-1)][i] 
	    + k*dt*dt/dx/dx
              *(u[n(0)][i+1] - 2*u[n(0)][i] + u[n(0)][i-1])
	    - (k-1)*dt*dt/dx
		*((1+du[i])/ds[i] - (1+du[i-1])/ds[i-1]);
	     // *(-.5)*(dv[i]*dv[i] - dv[i-1]*dv[i-1]));
	v[n(+1)][i] =  2*v[n(0)][i] - v[n(-1)][i] 
	    + dt*dt/dx/dx
		*(v[n(0)][i+1] - 2*v[n(0)][i] + v[n(0)][i-1])
            + (k-1)*dt*dt/dx
		*((ds[i]-1)/ds[i]*dv[i] - (ds[i-1]-1)/ds[i-1]*dv[i-1]);
             // *(du[i]*dv[i] - du[i-1]*dv[i-1] + 
	     //  .5*(dv[i]*dv[i]*dv[i] - dv[i-1]*dv[i-1]*dv[i-1]));
	if (violin) 
	    v[n(+1)][i] +=
		- .5*dt*dt*(fabs(i - IMAX/10) < 3)
		    *(fabs((v[n(+1)][i]-v[n(-1)][i])/(2*dt) + v0/3) < 2*v0/3);
    }
    u[n(+1)][0] = u[n(+1)][IMAX] = 0;
    v[n(+1)][0] = v[n(+1)][IMAX] = 0;

    if (n++ % draw_step ==0) glutPostRedisplay();
}

/*****************************************
 * display solution
 ----------------------------------------*/
void draw(double *u, double *v)
//
#define X(i)   ((double)width/IMAX*(i) )
#define Y(y)   ( (.25*(y)*ymag + .65)*height )
#define Y1(y)  ( (.25*(y)*ymag + .3)*height )
{
    glClearColor(.0,.0,.0,.0);
    glClear(GL_COLOR_BUFFER_BIT);

    /* axis */
    glColor3f(.2,.2,.3);
    glBegin(GL_LINES);
    glVertex2d(0, Y(0)); glVertex2d(width, Y(0));
    glVertex2d(0, Y1(0)); glVertex2d(width, Y1(0));
    glEnd();
    
    /* sol */
    glColor3f(.0,.9,.0);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i)+width*u[i],Y(v[i]));
    glEnd();

    glColor3f(.0,.6,.4);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i), Y1(u[i]*5));
    glEnd();

    if (violin) {
        static int count = 0;
        glColor3f(.6,.0,.3);
        glBegin(GL_LINES);
        double l = .004;
        for (int j = -10; j < 10; j++) {
            glVertex2d(X(IMAX/10), Y( l/4 + j*l - count*v0*dt));
            glVertex2d(X(IMAX/10), Y(-l/4 + j*l - count*v0*dt));
        }
        count = (count + draw_step) % int(l/v0/dt);
        glEnd();
    }
}

/*****************************************
 * glut call-back
 ----------------------------------------*/
void display(void) 
{
    draw(&u[0][0], &v[0][0]);
    if (dump_images) { /* to make gif-animation */
        static int count = 0;
	static int step = 20;
	glReadBuffer(GL_FRONT);
        if (count % step == 0) {
            char str[100];
            sprintf(str, "nlw%03d.bmp", count/step);
            SaveImageAs(str);
        }
	count++;
	if (count/step > 32) exit(0);
    }
    glutSwapBuffers();
}

void idle(void) { if (!pause) pde(); }

void reshape(int w, int h)
{
    glutReshapeWindow(width, (height = (h/2)*2));
    glViewport(0,0,width, height);
    glMatrixMode(GL_PROJECTION); {
	glLoadIdentity();
	gluOrtho2D(0., width - 1., 0., height - 1.);
    }
    pause = 0;
    glutPostRedisplay();
}

void action(int key)
{
    static const int kval[] = {1,2,4,9,20,50,100,200};
    switch (key) {
    case'q':  exit(0);
    case'c':  pause = !pause; break;
    case'1': case '2': case '3': case '4': case '5':
    case'6': case '7': case '8':
	k = kval[(key-'1')]; initialize=1; break;
    case'p': ivalue=key; initialize=1; break;
    case'g': ivalue=key; initialize=1; break;
    case'v': ivalue=key; initialize=1; break;
    default:  break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)  { action((int) key); }
void mouse(int state, int button, int x, int y) { action('c'); }

void make_menu(void)
{
    int kval_menu, sim_inst;
    kval_menu = glutCreateMenu(action); {
        glutAddMenuEntry("  1", '1');
        glutAddMenuEntry("  2", '2');
        glutAddMenuEntry("  4", '3');
        glutAddMenuEntry("  9", '4');
        glutAddMenuEntry(" 20", '5');
        glutAddMenuEntry(" 50", '6');
        glutAddMenuEntry("100", '7');
	glutAddMenuEntry("200", '8');
    }
    sim_inst = glutCreateMenu(action); {
        glutAddMenuEntry(" Piano",  'p');
        glutAddMenuEntry(" Guitar", 'g');
        glutAddMenuEntry(" Violin", 'v');
    }
    glutCreateMenu(action); {
        glutAddMenuEntry("Pause/Continue", 'c');
	glutAddSubMenu("k/T = ",     kval_menu);
	glutAddSubMenu("Simulate --", sim_inst);
        glutAddMenuEntry("Quit",           'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*****************************************
 * main
 ----------------------------------------*/
int main(int argc, char** argv)
{
    for (int idx = 1; idx < argc; idx++) {
        if (strcmp(argv[idx], "-b") == 0) { dump_images = 1; continue; }
        if (strcmp(argv[idx], "-p") == 0) { ivalue = 'p'; continue; }
        if (strcmp(argv[idx], "-g") == 0) { ivalue = 'g'; continue; }
        if (strcmp(argv[idx], "-v") == 0) { ivalue = 'v'; continue; }
        if (strcmp(argv[idx], "-k") == 0) {
	    sscanf(argv[++idx], "%lf", &k); continue; 
	}
        /* else: show usage */
        fprintf(stderr,
                "-k <kval>: assign k-value\n"
                "-b: dump each image to bmp-file\n\n"
                );
        exit(1);
    }

    /* init glut */
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);

    /* call-back */
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    make_menu();

    /* 2D-setup */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0., width - 1., 0., height - 1.);

    glutMainLoop ();
    return 0; 
}
