/*
 * 1-dim Wave equation: u_tt - u_xx = 0
 */

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>

#define PI 3.14159265358979323846

/* global staff */
#define IMAX 500
#define NBUF 3
double u[NBUF][IMAX+1];

/* window size */
int width = IMAX, height = width/2;

/* solving equation */
void Wave(void);
void (*pde)(void) = Wave;

/* global parameters */
#define L   (1.0)   /* length */
int ivalue = 0;

/* flags */
int initialize = 1;
int pause = 0;

/* for violin */
int violin = 0; 
#define v0  (.003) // velocity of "yumi"

/* drawing parameters */
int draw_step = 8;
double  ymag  = 100.0;

/*****************************************
 * solve Wave equation
 ----------------------------------------*/
void Wave(void)
//
#define dx   ( L/IMAX )
#define x(i) ( dx*(i) )
#define dt   ( dx )
#define n(D) ( (n+(D)+NBUF) % NBUF )
#define sch1 ( 1./cosh(30*(x(i) - .5)) )
#define sch2 ( 1./cosh(30*(x(i) - .6)) )
{
    static unsigned int n = 0;
    /* initial value */
    if (initialize == 1) {
	double vmax = .01;
	n = 0; initialize = 0; violin = 0;
	switch (ivalue) {
	case 'g':
	    for (int i = 1; i < IMAX; i++) {
		static double x0 = L/6;
		u[n(0)][i] = u[n(-1)][i] = 
		    - vmax*((x(i)<x0) ? x(i)/x0 : (L-x(i))/(L-x0));
	    }
	    ivalue = 0; break;
	case 'p':
	    for (int i = 1; i < IMAX; i++) {
		static double x0 = L/8;
		u[n(0)][i]  = 0;
		u[n(-1)][i] = -.3*dt*(1./cosh(60*(x(i) - x0)));
	    }
	    ivalue = 0; break;
	case 'v':
	    for (int i = 1; i < IMAX; i++) 
		u[n(0)][i] = u[n(-1)][i] = 0;
	    violin =1;
	    ivalue = 0; break;
	case 'r':
	    for (int i = 1; i < IMAX; i++) 
		u[n(0)][i] = u[n(-1)][i+1] = vmax*(sch1-.5*sch2);
	    ivalue = 0; break;
	case 'd':
	    for (int i = 1; i < IMAX; i++) 
		u[n(0)][i] = u[n(-1)][i] = vmax*1.5*sch1;
	    ivalue = 0; break;
	default:
	    for (int i = 1; i < IMAX; i++) {
		u[n(0)][i]=u[n(-1)][i]= 
		    vmax*(  ((ivalue&(1<<0))!=0)*sin( 1*PI*x(i))*.9
			  + ((ivalue&(1<<1))!=0)*sin( 2*PI*x(i))*.8
			  + ((ivalue&(1<<2))!=0)*sin( 3*PI*x(i))*.7
			  + ((ivalue&(1<<3))!=0)*sin( 4*PI*x(i))*.6
			  + ((ivalue&(1<<4))!=0)*sin( 5*PI*x(i))*.5
			  + ((ivalue&(1<<5))!=0)*sin( 6*PI*x(i))*.4
			  + ((ivalue&(1<<6))!=0)*sin( 8*PI*x(i))*.3
			  + ((ivalue&(1<<7))!=0)*sin(10*PI*x(i))*.2
			  + ((ivalue&(1<<8))!=0)*sin(20*PI*x(i))*.2 );
	    }
		break;
	}
    }

    /* evolve */
    for (int i = 1; i < IMAX; i++) {
	u[n(+1)][i] = 2*u[n(0)][i] - u[n(-1)][i] 
	    + dt*dt/dx/dx*(u[n(0)][i+1] - 2*u[n(0)][i] + u[n(0)][i-1]);
        if (violin) 
            u[n(+1)][i] +=
		- .5*dt*dt
		    *(fabs(i - IMAX/10) < 3)
		   *(fabs((u[n(+1)][i]-u[n(-1)][i])/(2*dt) + v0/3) < 2*v0/3);
    }
    u[n(+1)][0] = u[n(+1)][IMAX] = 0;
    // u[n(+1)][IMAX] = u[n(0)][IMAX-1]; /* path through */

    if (n++ % draw_step ==0) glutPostRedisplay();
}

/*****************************************
 * display solution
 ----------------------------------------*/
void draw(double *fn)
//
#define X(i)   ((double)width/IMAX*(i) )
#define Y(y)   ( (.25*(y)*ymag + .5)*height )
{
    glClearColor(.0,.0,.0,.0);
    glClear(GL_COLOR_BUFFER_BIT);

    /* axis */
    glColor3f(.2,.2,.3);
    glBegin(GL_LINES);
    glVertex2d(0, Y(0)); glVertex2d(width, Y(0));
    glEnd();
    
    /* sol */
    glColor3f(.0,.9,.0);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i <= IMAX; i++)
	glVertex2d(X(i),Y(fn[i]));
    glEnd();
    if (violin) {
        static int count = 0;
        glColor3f(.6,.0,.3);
	glBegin(GL_LINES);
	double l = .004;
	for (int j = -10; j < 10; j++) {
	    glVertex2d(X(IMAX/10), Y( l/4 + j*l - count*v0*dt));
	    glVertex2d(X(IMAX/10), Y(-l/4 + j*l - count*v0*dt));
	}
	count = (count + draw_step) % int(l/v0/dt);
	glEnd();
    }
}

/*****************************************
 * glut call-back
 ----------------------------------------*/
void display(void) 
{
    draw(&u[0][0]);
    glutSwapBuffers();
}

void idle(void) { if (!pause) pde(); }

void reshape(int w, int h)
{
    glutReshapeWindow(width, (height = (h/2)*2));
    glViewport(0,0,width, height);
    glMatrixMode(GL_PROJECTION); {
	glLoadIdentity();
	gluOrtho2D(0., width - 1., 0., height - 1.);
    }
    pause = 0;
    glutPostRedisplay();
}

void action(int key)
{
    switch (key) {
    case'q':  exit(0);
    case'c':  pause = !pause; break;
    case'*': draw_step=draw_step*2; break;
    case'/': draw_step=draw_step/2 + 1; break;
    case'1': case '2': case '3': case '4': case '5':
    case'6': case '7': case '8': case '9':
	ivalue^=(1<<(key-'1')); initialize=1; break;
    case'0': ivalue=  0; initialize=1; break;
    case'p': ivalue=key; initialize=1; break;
    case'r': ivalue=key; initialize=1; break;
    case'd': ivalue=key; initialize=1; break;
    case's': ivalue=key; initialize=1; break;
    case'g': ivalue=key; initialize=1; break;
    case'v': ivalue=key; initialize=1; break;
    default:  break;
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)  { action((int) key); }
void mouse(int state, int button, int x, int y) { action('c'); }

void make_menu(void)
{
    int mode_menu, sim_inst;
    mode_menu = glutCreateMenu(action); {
        glutAddMenuEntry("Clear", '0');
        glutAddMenuEntry(" 1", '1');
        glutAddMenuEntry(" 2", '2');
        glutAddMenuEntry(" 3", '3');
        glutAddMenuEntry(" 4", '4');
        glutAddMenuEntry(" 5", '5');
        glutAddMenuEntry(" 6", '6');
        glutAddMenuEntry(" 8", '7');
        glutAddMenuEntry("10", '8');
        glutAddMenuEntry("20", '9');
    }
    sim_inst = glutCreateMenu(action); {
        glutAddMenuEntry(" Piano",  'p');
        glutAddMenuEntry(" Guitar", 'g');
        glutAddMenuEntry(" Violin", 'v');
    }
    glutCreateMenu(action); {
        glutAddMenuEntry("Pause/Continue", 'c');
        glutAddMenuEntry("Reflection",     'r');
	glutAddSubMenu("Toggle mode", mode_menu);
	glutAddSubMenu("Simulate --", sim_inst);
        glutAddMenuEntry("Slow down",      '/');
        glutAddMenuEntry("Speed up",       '*');
        glutAddMenuEntry("Quit",           'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/*****************************************
 * main
 ----------------------------------------*/
int main(int argc, char** argv)
{
    /* init glut */
    glutInit (&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(argv[0]);

    /* call-back */
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(keyboard);
    glutMouseFunc(mouse);
    make_menu();

    /* 2D-setup */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0., width - 1., 0., height - 1.);

    glutMainLoop ();
    return 0; 
}
